import { AsyncLocalStorage } from 'async_hooks';

class ScopeContextError extends Error {
}
const scoped = (ClassCtor) => {
    const asyncStorage = new AsyncLocalStorage();
    const referenceMap = new WeakMap();
    class ClassReferer {
        constructor() {
            const proxyInstance = new Proxy(this, {
                get(_, propKey, receiver) {
                    if (propKey === 'init') {
                        return;
                    }
                    const referenceKey = asyncStorage.getStore();
                    if (!referenceKey) {
                        throw new ScopeContextError('di-scoped ContextReferer not running in context');
                    }
                    let reference = referenceMap.get(referenceKey);
                    if (!reference) {
                        reference = referenceMap.set(referenceKey, new ClassCtor(...referenceKey)).get(referenceKey);
                    }
                    return Reflect.get(reference, propKey, receiver);
                },
                set(_, propKey, value, receiver) {
                    const referenceKey = asyncStorage.getStore();
                    if (!referenceKey) {
                        throw new ScopeContextError('di-scoped ContextReferer not running in context');
                    }
                    let reference = referenceMap.get(referenceKey);
                    if (!reference) {
                        reference = referenceMap.set(referenceKey, new ClassCtor(...referenceKey)).get(referenceKey);
                    }
                    return Reflect.set(reference, propKey, value, receiver);
                },
            });
            Object.setPrototypeOf(this, proxyInstance);
        }
    }
    const classInstance = new ClassReferer();
    function ClassActivator() {
        return classInstance;
    }
    ClassActivator.runInContext = (fn, ...args) => {
        referenceMap.set(args, new ClassCtor(...args));
        return asyncStorage.run(args, fn);
    };
    ClassActivator.runOutOfContext = (fn) => {
        return () => asyncStorage.run(undefined, fn);
    };
    ClassActivator.runAsyncIterator = function (iterator, ...ctorArgs) {
        referenceMap.set(ctorArgs, new ClassCtor(...ctorArgs));
        return {
            async next(...args) {
                return asyncStorage.run(ctorArgs, () => iterator.next(...args));
            },
            async return(value) {
                if (iterator.return) {
                    return asyncStorage.run(ctorArgs, () => iterator.return(value));
                }
                return Promise.resolve({ value: undefined, done: true });
            },
            async throw(...args) {
                if (iterator.throw) {
                    return asyncStorage.run(ctorArgs, () => iterator.throw(...args));
                }
                return Promise.reject(new Error("Iterator does not support throwing errors"));
            },
            [Symbol.asyncIterator]() {
                return this;
            },
            // @ts-ignore
            async [Symbol.asyncDispose]() {
                // @ts-ignore
                if (typeof iterator[Symbol.asyncDispose] === "function") {
                    // @ts-ignore
                    return asyncStorage.run(ctorArgs, () => iterator[Symbol.asyncDispose]());
                }
                return Promise.resolve();
            },
        };
    };
    ClassActivator.runIterator = function (generator, ...ctorArgs) {
        referenceMap.set(ctorArgs, new ClassCtor(...ctorArgs));
        return {
            next(...args) {
                return asyncStorage.run(ctorArgs, () => generator.next(...args));
            },
            return(value) {
                if (generator.return) {
                    return asyncStorage.run(ctorArgs, () => generator.return(value));
                }
                return { value: undefined, done: true };
            },
            throw(...args) {
                if (generator.throw) {
                    return asyncStorage.run(ctorArgs, () => generator.throw(...args));
                }
                throw new Error("Generator does not support throwing errors");
            },
            [Symbol.iterator]() {
                return this;
            },
        };
    };
    ClassActivator.hasContext = () => {
        return !!asyncStorage.getStore();
    };
    return ClassActivator;
};
/*
const TestClass = scoped(class {

    constructor(private name: string) {
    }

    test() {
      return `Hello, ${this.name}`;
    }
});
*/
/*
const sleep = () => new Promise<void>((res) => setTimeout(() => res(), 100))

TestClass.runInContext(async () => {
  await sleep()
  await TestClass.runOutOfContext(async () => {
    await sleep()
    await TestClass.runInContext(async () => {
      await sleep()
      console.log(new TestClass().test())
    }, "foo2");
  })

}, "foo1");
*/
/*
class TestIterator {

  test = new TestClass();

  *run() {
    yield this.test.test()
    yield this.test.test()
    yield this.test.test()
  }
}

const test = new TestIterator()

console.log([...TestClass.runIterator(test.run(), "Pahom")])
*/
// TestClass.runInContext(() => {
//    new TestClass().test(); // Hello, Peter
// }, "Peter")
// new TestClass().test(); // ScopeContextError('di-scoped ContextReferer not running in context');
/*
let instanceRef1;
let instanceRef2;

TestClass.runInContext(() => {
    instanceRef1 = new TestClass()
    instanceRef1.test() // Hello, Peter
}, "Peter")

TestClass.runInContext(() => {
    instanceRef2 = new TestClass()
    instanceRef2.test() // Hello, not Peter
}, "not Peter")

if (TestClass === TestClass) {
    console.log("Ok! This is the same class")
}
if (instanceRef1 === instanceRef2) {
    console.log("OMG! This is the same instance")
}
*/

export { ScopeContextError, scoped };
